{# Import the theme's layout. #}
{% extends "!layout.html" %}

{# Add some extra stuff before and use existing with 'super()' call. #}

{%- block footer %}
    <div class="footer">
        <div class="container">
                    &copy;  {{ copyright }}
                        <a href="http://github.com/hobu">Howard Butler</a>,
                        <a href="http://github.com/mpgerlek">Michael Gerlek</a>,
                        and
                        <a href="https://github.com/PDAL/PDAL/graphs/contributors">others</a>,
            {%- if last_updated %}
                {% trans last_updated=last_updated|e %}Last updated
                    on {{ last_updated }}.{% endtrans %}
            {%- endif %}
        </div>
    </div>
{%- endblock %}
