{% if theme_coursename %}
  <div class="rst-versions" data-toggle="rst-versions" role="note" aria-label="versions">
    <span class="rst-current-version" data-toggle="rst-current-version">
      <span class="fa fa-book"><a href="{{ pathto(master_doc) | replace('index.html','') | replace('#','') }}../index.html"> {{ theme_coursename|striptags }} </a></span>
    </span>
  </div>
{% endif %}

