<footer>
  {% if (theme_prev_next_buttons_location == 'bottom' or theme_prev_next_buttons_location == 'both') and (next or prev) %}
    <div class="rst-footer-buttons" role="navigation" aria-label="footer navigation">
      {% if next %}
        <a href="{{ next.link|e }}" class="btn btn-neutral float-right" title="{{ next.title|striptags|e }}" accesskey="n" rel="next">{{ _('Next') }} <span class="fa fa-arrow-circle-right"></span></a>
      {% endif %}
      {% if prev %}
        <a href="{{ prev.link|e }}" class="btn btn-neutral float-left" title="{{ prev.title|striptags|e }}" accesskey="p" rel="prev"><span class="fa fa-arrow-circle-left"></span> {{ _('Previous') }}</a>
      {% endif %}
    </div>
  {% endif %}

  <hr/>

  <div role="contentinfo">
    <div class="info">
      <a class="logo-link" href="https://hobu.co">
        <div class="hobu-logo"></div>
      </a>
      <div class="copyright">
      {% if copyright %}

                    &copy;  {{ copyright }}
                        <a href="https://github.com/abellgithub">Andrew Bell</a>,
                        <a href="https://github.com/chambbj">Brad Chambers</a>,
                        <a href="http://github.com/hobu">Howard Butler</a>,
                        and
                        <a href="https://github.com/PDAL/PDAL/graphs/contributors">others</a>.

      {% endif %}
   {%- if build_id and build_url %}
      {% trans build_url=build_url, build_id=build_id %}
        <span class="build">
          Build
          <a href="{{ build_url }}">{{ build_id }}</a>.
        </span>
      {% endtrans %}
    {%- elif commit %}
      {% trans commit=commit %}
        <span class="commit">
          Revision <code>{{ commit }}</code>.
        </span>
      {% endtrans %}
    {%- elif last_updated %}
      {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}
    {%- endif %}
      </div>
    </div>
</footer>
