/******************************************************************************
* Copyright (c) 2011, Michael P. Gerlek (mpg@flaxen.com)
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following
* conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in
*       the documentation and/or other materials provided
*       with the distribution.
*     * Neither the name of Hobu, Inc. or Flaxen Geo Consulting nor the
*       names of its contributors may be used to endorse or promote
*       products derived from this software without specific prior
*       written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
* COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
* AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
****************************************************************************/

#include <pdal/pdal_test_main.hpp>

#include "Support.hpp"
#include "io/BpfSupport.hpp"

using namespace pdal;


TEST(BpfTestZlib, test_point_major_zlib)
{
    test_file_type(
        Support::datapath("bpf/"
            "autzen-utm-chipped-25-v3-deflate-interleaved.bpf"));
}

TEST(BpfTestZlib, test_dim_major_zlib)
{
    test_file_type(
        Support::datapath("bpf/autzen-utm-chipped-25-v3-deflate.bpf"));
}

TEST(BpfTestZlib, test_byte_major_zlib)
{
    test_file_type(
        Support::datapath("bpf/"
            "autzen-utm-chipped-25-v3-deflate-segregated.bpf"));
}

TEST(BpfTestZlib, roundtrip_byte_compression)
{
    Options ops;

    ops.add("format", "BYTE");
    ops.add("compression", true);
    test_roundtrip(ops);
}

TEST(BpfTestZlib, roundtrip_dimension_compression)
{
    Options ops;

    ops.add("format", "DIMENSION");
    ops.add("compression", true);
    test_roundtrip(ops);
}

TEST(BpfTestZlib, roundtrip_point_compression)
{
    Options ops;

    ops.add("format", "POINT");
    ops.add("compression", true);
    test_roundtrip(ops);
}

